from geant4_pybind import *
from g4camp.PrimaryGeneratorAction import PrimaryGeneratorAction
from g4camp.RunAction import RunAction
from g4camp.EventAction import EventAction
from g4camp.StackingAction import StackingAction
from g4camp.SteppingAction import SteppingAction
from g4camp.DataBuffer import DataBuffer

class ActionInitialization(G4VUserActionInitialization):

    def __init__(self, app, primary_generator, gun_args={}):
        super().__init__()
        self.app = app
        self.primary_generator = primary_generator
        self.gun_args = gun_args

    def BuildForMaster(self):  # invoked in multithread mode only
        self.SetUserAction(RunAction(True))

    def Build(self):           # invoked in in boths modes: multihread - multiple time, serial - once
        self.primGenAct = PrimaryGeneratorAction(self.primary_generator, gun_args=self.gun_args)
        self.SetUserAction(self.primGenAct)
        self.SetUserAction(RunAction(False))
        self.SetUserAction(EventAction(self.app))
        self.SetUserAction(StackingAction(self.app))
        self.SetUserAction(SteppingAction(self.app))



from g4camp.g4camp import g4camp
import time
import numpy as np
import h5py
import sys
import configargparse
import logging

log = logging.getLogger('run_g4camp')
logformat='[%(name)12s ] %(levelname)8s: %(message)s'
logging.basicConfig(format=logformat)

def report_timing(func):
    def func_wrapper(*args, **kwargs):
        tic = time.time()
        result = func(*args, **kwargs)
        toc = time.time()
        dt = toc - tic
        log .info(f'{func.__name__}'.ljust(40)+f'{dt:6.3f} seconds'.rjust(30))
        return result
    return func_wrapper


def parse_arguments():
    p = configargparse.get_argument_parser()
    p.add_argument('-n', default=100, help="number of events")
    p.add_argument('-o', default='output.h5', help="output file name")
    p.add_argument('--enable_optics', help="enable optics", action='store_true')
    p.add_argument('-p', '--primary-generator', default='gun', choices=('gun', 'gps'), help="type of primary generator type")
    p.add_argument('--gps_macro', default='muons.mac', help="macro files with Geant4 GPS commands to congigure particle source")
    p.add_argument('--skip_min', default=0.002, help="minimal of particle energy to skip")
    p.add_argument('--skip_max', default=0.01, help="maximal of particle energy to skip")
    p.add_argument('--random_seed', default=1, help="random seed")
    p.add_argument('--photon_suppression', default=10, help="photon suppression factor")
    p.add_argument('-l', '--log-level', choices=('deepdebug', 'debug', 'info', 'warning', 'error', 'critical'), default='info', help='logging level')

    opts = p.parse_args()
    #
    log.setLevel(opts.log_level.upper())
    log.info("----------")
    log.info(p.format_help())
    log.info("----------")
    log.info(p.format_values())    # useful for logging where different settings came from
    return opts


@report_timing
def run(app, n_events=10, ofname="output.h5"):
    h5file = h5py.File(ofname, "w")  # create empty file
    h5file.close()
    for ievt, data in enumerate(app.run(n_events)):
        h5file = h5py.File(ofname, "a")
        g = h5file.create_group(f"event_{ievt}")
        g.create_dataset('photons', data = data.photon_cloud)
        vpositions = [vertex[0] for vertex in data.vertices]
        vparticles = [vertex[2][0] for vertex in data.vertices]
        g.create_dataset('vertex_positions', data = vpositions)
        g.create_dataset('vertex_particles', data = vparticles)
        #g.create_dataset('tracks', data = data.tracks)
        g.create_dataset('tracks', 
                          data = np.array( [tuple(row) for row in data.tracks], 
                                           dtype = [('uid', float), ('pdgid', float), 
                                                    ('x_m', float), ('y_m', float), ('z_m', float),
                                                    ('t_ns', float),('E_GeV', float)]) 
                        )
        h5file.close()
        #log.info(f" Event #{ievt}/{n_events}")
        log.info(f"   Number of vertices:     {len(data.vertices)}")
        log.info(f"   Number of tracks:       {len(np.unique(data.tracks[:,0]))}")
        log.info(f"   Number of track points: {len(data.tracks)}")
        log.info(f"   Number of photons:      {len(data.photon_cloud)}")
        db = app.data_buffer
        #log.info(f"   Number of vertices / tracks points / photons : {len(db.vertices):6.0f} / {len(db.tracks):6.0f} / {len(db.photon_cloud):6.0f}")
        log.info(" ")


def main():
    opts = parse_arguments()
    optics = opts.enable_optics
    #
    if opts.primary_generator == 'gun':
        gun_args = {'particle': 'mu-', 
                    'energy_GeV': 200, 
                    'position_m': [0,0,0],
                    'direction': [0,0,1]}
        app = g4camp(optics=optics, primary_generator='gun', gun_args=gun_args)
    elif opts.primary_generator == 'gps':
        app = g4camp(optics=optics, primary_generator='gps')
        app.setGPSMacro(opts.gps_macro)
    else:
        log.error(f"Wrong primary type '{opts.primary_generator}'")
    app.log = log
    app.setRandomSeed(int(opts.random_seed))
    app.setSkipMinMax(float(opts.skip_min), float(opts.skip_max))
    app.setPhotonSuppressionFactor(float(opts.photon_suppression))
    app.configure()
    run(app, n_events=int(opts.n), ofname=opts.o)


if __name__ == "__main__":
    main()




from geant4_pybind import *
#from MuNuclearBuilder import MuNuclearBuilder

class CustomPhysicsList(G4VModularPhysicsList):

  def __init__(self, optics=True):
    super().__init__()

    self.phys_constructors = {}

    self.phys_constructors["decay"] = G4DecayPhysics()
    self.phys_constructors["em"] = G4EmStandardPhysics_option1()   # G4EmStandardPhysics(),  G4EmStandardPhysics_option4()
    self.phys_constructors["em_extra"] = G4EmExtraPhysics()  # for muon nuclear interaction
#    self.phys_constructors["rad_decay"] = G4RadioactiveDecayPhysics()
    
    if optics:
      self.phys_constructors["optical"] = G4OpticalPhysics()

    for key, phys in self.phys_constructors.items():
      self.RegisterPhysics(phys)
      phys.SetVerboseLevel(0)




from geant4_pybind import *

class PrimaryGeneratorAction(G4VUserPrimaryGeneratorAction):

    def __init__(self, primary_generator, gun_args={}):
        super().__init__()
        if primary_generator == 'gps':
            self.particleGun = G4GeneralParticleSource()
        elif primary_generator == 'gun':
            nofParticles = 1
            self.particleGun = G4ParticleGun(nofParticles)
            gun_args = self.FillMissingGunParameters(gun_args)
            particle_table = G4ParticleTable.GetParticleTable()
            particleDefinition = particle_table.FindParticle(gun_args['particle'])
            self.particleGun.SetParticleDefinition(particleDefinition)
            self.particleGun.SetParticleMomentumDirection(G4ThreeVector(*gun_args['direction']))
            self.particleGun.SetParticlePosition(G4ThreeVector(*gun_args['position_m']))
            self.particleGun.SetParticleEnergy(gun_args['energy_GeV'])
        else:
            sys.exit("Wrong primary generator type ({primary_generator}), it can be either 'gps' or 'gun'")


    def FillMissingGunParameters(self, gun_args):
        if 'particle' not in gun_args.keys():
            gun_args['particle'] = 'mu-'
        if 'energy_GeV' not in gun_args.keys():
            gun_args['energy_GeV'] = 100*GeV
        if 'position_m' not in gun_args.keys():
            gun_args['position_m'] = [0, 0, 0]
        if 'direction' not in gun_args.keys():
            gun_args['direction'] = [0, 0, 1]
        return gun_args
        

    def GeneratePrimaries(self, anEvent):
        # This function is called at the begining of event
        self.particleGun.GeneratePrimaryVertex(anEvent)




from geant4_pybind import *
import numpy as np

class DetectorConstruction(G4VUserDetectorConstruction):

  def __init__(self):
    super().__init__()
    self.rho = 1500*m
    self.zmin = 0*m
    self.zmax = 1500.*m
    self.rock_depth = 200.*m
    self.check_overlaps = True
    #
    #self.mie_g_forward = 0.98
    #self.mie_g_backward = 0.99
    #self.mie_f_to_b_ratio = 1.0
    #self.mie_fraction = 0.99
    #
    self.sensDet = None
    #
    self.DefineMaterials()


  def DefineMaterials(self):
    man = G4NistManager.Instance()
    elH  = man.FindOrBuildElement("H", False)
    elO  = man.FindOrBuildElement("O", False)
    elC  = man.FindOrBuildElement("C", False)
    elCa = man.FindOrBuildElement("Ca", False)
    elMg = man.FindOrBuildElement("Mg", False)
    #
    # Water
    nelements = 2
    self.matWater = G4Material("Water", 1.0*g/cm3, nelements, kStateLiquid)
    self.matWater.AddElement(elH, 2)
    self.matWater.AddElement(elO, 1)
    #
    # Rock for lake bed
    nelements = 4
    self.matStdRock = G4Material("StdRock", 2.65*g/cm3, nelements, kStateSolid );
    self.matStdRock.AddElement(elO,  0.52);
    self.matStdRock.AddElement(elCa, 0.27);
    self.matStdRock.AddElement(elC,  0.12);
    self.matStdRock.AddElement(elMg, 0.09);
    #
    # Optical properties
    #  (just some dummy values)
    ph_ene = 1239.84193/np.array([610, 350])*eV
    ph_rindex = np.full(ph_ene.shape, fill_value=1.34)
    ph_abs =    np.full(ph_ene.shape, fill_value=20*m)
    ph_scat =   np.full(ph_ene.shape, fill_value=60*m)

    MPT_water = G4MaterialPropertiesTable()
    MPT_water.AddProperty("RINDEX", G4doubleVector(ph_ene), G4doubleVector(ph_rindex))
    MPT_water.AddProperty("ABSLENGTH", G4doubleVector(ph_ene), G4doubleVector(ph_abs))
    MPT_water.AddProperty("RAYLEIGH", G4doubleVector(ph_ene), G4doubleVector(ph_scat))
    #MPT_water.AddProperty("MIEHG", G4doubleVector(ph_ene), G4doubleVector(ph_scat))
    #MPT_water.AddConstProperty("MIEHG_FORWARD", self.mie_g_forward);
    #MPT_water.AddConstProperty("MIEHG_BACKWARD", self.mie_g_backward);
    #MPT_water.AddConstProperty("MIEHG_FORWARD_RATIO", self.mie_f_to_b_ratio)
    self.matWater.SetMaterialPropertiesTable(MPT_water)


  def Construct(self):
    # Get materials
    matManager = G4NistManager.Instance()
    matWater = matManager.FindOrBuildMaterial("Water");
    matStdRock = matManager.FindOrBuildMaterial("StdRock");
    #
    # Mother volume
    worldRho = 1.1 * self.rho
    worldH   = 1.1 * 2. * max(abs(self.zmax),abs(self.zmin))
    world_solid = G4Tubs("World", 0., worldRho, worldH/2., 0., 360.*deg)
    world_log   = G4LogicalVolume(world_solid, matWater, "World")
    world_phys  = G4PVPlacement(None, G4ThreeVector(), world_log,
                                "World", None, False, self.check_overlaps)
    #
    # Lake bed
    lakeBed_solid = G4Tubs("LakeBed", 0., self.rho, self.rock_depth/2., 0, 360*deg)
    lakeBed_log   = G4LogicalVolume(lakeBed_solid, matStdRock, "LakeBed")
    G4PVPlacement(None, G4ThreeVector(0,0,-self.rock_depth/2.), lakeBed_log,
                  "LakeBed", world_log, False, self.check_overlaps);
    #
    # Segmented sensitive volume
    water_solid = G4Tubs("Water", 0., self.rho, (self.zmax-self.zmin)/2., 0, 360*deg)
    water_log   = G4LogicalVolume(water_solid, matWater, "Water")
    G4PVPlacement(None, G4ThreeVector(0,0,(self.zmax+self.zmin)/2.), water_log,
                  "Water", world_log, False, self.check_overlaps);
    self.det_log = water_log
    #

    return world_phys


  def ConstructSDandField(self):
    pass
    #self.sensDet = SensitiveDetector("SensDet")
    #G4SDManager.GetSDMpointer().AddNewDetector(self.sensDet)
    #self.det_log.SetSensitiveDetector(self.sensDet)


  def GetSensitiveDetector(self):
    return self.sensDet



import numpy as np

class DataBuffer():

  def __init__(self):
    self.particles_per_bunch = 10000
    self.particle_bunches = 1
    self.particle_num = 0
    self.particles = np.zeros((self.particles_per_bunch, 9), dtype=float)
    # 'particle' row: pdgid, x[m], y[m], z[m], t[ns], Px[GeV], Py[GeV], Pz[GeV], Etot[GeV]
    #
    self.track_points_per_bunch = 100000
    self.track_point_bunches = 1
    self.point_num = 0
    self.tracks = np.zeros((self.track_points_per_bunch, 7), dtype=float) 
    # 'track' row:  uid, pdgid, x[m], y[m], z[m], t[ns], Etot[GeV]
    self.photon_cloud = []
    self.status = 'empty' # 'empty' -> 'in_progress' -> 'ready'

  def AddParticle(self, pdgid, x, y, z, t, Px, Py, Pz, Etot):
    self.status = 'in_progress'
    self.particles[self.particle_num] = [pdgid, x, y, z, t, Px, Py, Pz, Etot]
    self.particle_num += 1
    if self.particle_num % self.particles_per_bunch == 0:
      self.particles = np.concatenate((self.particles, np.zeros((self.particles_per_bunch, 9), dtype=int)))
      self.particle_bunches += 1

  def AddPhoton(self, photon):
    self.status = 'in_progress'
    self.photon_cloud.append(photon)

  def AddTrackPoint(self, uid, pdgid, x, y, z, t, Etot):
    self.status = 'in_progress'
    self.tracks[self.point_num] = [uid, pdgid, x, y, z, t, Etot]
    self.point_num += 1
    if self.point_num % self.track_points_per_bunch == 0:
      self.tracks = np.concatenate((self.tracks, np.zeros((self.track_points_per_bunch, 7), dtype=int)))
      self.track_point_bunches += 1

  def CutEmptyItems(self):
    self.tracks = self.tracks[:self.point_num]
    self.particles = self.particles[:self.particle_num]

  def SortItems(self):
    self.tracks = self.tracks[np.lexsort((self.tracks[:,5], self.tracks[:,0]))]

  def Clear(self):
    self.__init__()

  def Close(self):
    self.status = 'ready'

  def IsEmpty(self):
    return self.status == 'empty'

  def IsReady(self):
    return self.status == 'ready'

